package mccity.heroes.skills.rockslide;

import com.herocraftonline.heroes.Heroes;
import com.herocraftonline.heroes.api.SkillResult;
import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import com.herocraftonline.heroes.characters.skill.SkillType;
import com.herocraftonline.heroes.characters.skill.TargettedSkill;
import com.herocraftonline.heroes.util.Messaging;
import com.herocraftonline.heroes.util.Setting;
import com.herocraftonline.heroes.util.Util;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.craftbukkit.entity.CraftPlayer;
import org.bukkit.craftbukkit.event.CraftEventFactory;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.inventory.ItemStack;

import java.util.*;

public class SkillRockslide extends TargettedSkill {

    private static final EnumSet<Material> FALL_FREE_MATS = EnumSet.of(Material.AIR, Material.WATER, Material.STATIONARY_WATER,
            Material.LAVA, Material.STATIONARY_LAVA);

    // Prevents Gravel generator - spamming rockslide at the same location
    // Center location -> radius
    private final Map<Location, Integer> currentRockslides = new HashMap<Location, Integer>();

    private String useFailedText;

    public SkillRockslide(Heroes plugin) {
        super(plugin, "Rockslide");
        setDescription("Brings down a Rockslide on the target.");
        setUsage("/skill rockslide <target>");
        setArgumentRange(0, 1);
        setIdentifiers("skill rockslide");
        setTypes(SkillType.EARTH, SkillType.HARMFUL, SkillType.SILENCABLE);
    }

    @Override
    public void init() {
        super.init();
        useFailedText = SkillConfigManager.getRaw(this, "use-failed-text", "Cant use here");
    }

    public ConfigurationSection getDefaultConfig() {
        ConfigurationSection defaultConfig = super.getDefaultConfig();
        defaultConfig.set(Setting.COOLDOWN.node(), 30000);
        defaultConfig.set(Setting.MANA.node(), 30);
        defaultConfig.set(Setting.MAX_DISTANCE.node(), 10);
        defaultConfig.set(Setting.MAX_DISTANCE_INCREASE.node(), 0.1);
        defaultConfig.set(Setting.DURATION.node(), 2000);
        defaultConfig.set(Setting.DURATION_INCREASE.node(), 60);
        defaultConfig.set("radius", 1);
        defaultConfig.set("height", 3);
        defaultConfig.set("use-failed-text", "Cant use here");
        return defaultConfig;
    }

    public String getDescription(Hero hero) {
        StringBuilder descr = new StringBuilder(getDescription());

        double cdSec = SkillConfigManager.getUseSetting(hero, this, Setting.COOLDOWN, 30000, false) / 1000.0;
        if (cdSec > 0) {
            descr.append(" CD:");
            descr.append(Util.formatDouble(cdSec));
            descr.append("s");
        }

        int mana = SkillConfigManager.getUseSetting(hero, this, Setting.MANA, 30, false);
        if (mana > 0) {
            descr.append(" M:");
            descr.append(mana);
        }

        double distance = SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE.node(), 10, false) +
                SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE_INCREASE.node(), 0.1, false) * hero.getSkillLevel(this);
        if (distance > 0) {
            descr.append(" Dist:");
            descr.append(Util.formatDouble(distance));
        }

        return descr.toString();
    }

    @Override
    public SkillResult use(Hero hero, LivingEntity target, String[] args) {
        Location rockslideLoc = target.getLocation().getBlock().getLocation();
        rockslideLoc.add(0.0, Math.ceil(target.getEyeHeight()) + 1, 0.0);
        World world = rockslideLoc.getWorld();

        int radius = SkillConfigManager.getUseSetting(hero, this, "radius", 1, false);
        radius = Math.max(0, Math.min(radius, 3));

        if (!canBuild(hero.getPlayer(), rockslideLoc) || isRockslideCrossingOther(rockslideLoc, radius)) {
            Messaging.send(hero.getPlayer(), useFailedText);
            return SkillResult.FAIL;
        }

        int height = SkillConfigManager.getUseSetting(hero, this, "height", 3, false);
        height = Math.max(1, Math.min(height, 5));

        RemoveGravelTask rmTask = new RemoveGravelTask(rockslideLoc);

        int x = rockslideLoc.getBlockX();
        int y = rockslideLoc.getBlockY();
        int z = rockslideLoc.getBlockZ();

        for (int ix = x - radius; ix <= x + radius; ix++) {
            for (int iz = z - radius; iz <= z + radius; iz++) {
                for (int iy = y; iy < y + height; iy++) {
                    Location curLoc = new Location(world, ix, iy, iz);
                    Block curBlock = world.getBlockAt(curLoc);
                    if (curBlock.getType() == Material.AIR) {
                        rmTask.addGravel(curBlock);
                        curBlock.setType(Material.GRAVEL);
                        breakTorchesBelow(curBlock);
                    }
                }
            }
        }

        if (rmTask.notEmpty()) {
            currentRockslides.put(rockslideLoc, radius);
            int durationMs = SkillConfigManager.getUseSetting(hero, this, Setting.DURATION.node(), 2000, false) +
                    SkillConfigManager.getUseSetting(hero, this, Setting.DURATION_INCREASE.node(), 60, false) * hero.getSkillLevel(this);
            Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, rmTask, durationMs / 50);
        }

        return SkillResult.NORMAL;
    }

    private boolean canBuild(Player player, Location rockslideLoc) {
        Block rockSlideBlock = rockslideLoc.getWorld().getBlockAt(rockslideLoc);
        if (rockSlideBlock.getType() != Material.AIR) {
            return false;
        }

        BlockPlaceEvent bpe = CraftEventFactory.callBlockPlaceEvent(((CraftWorld) rockslideLoc.getWorld()).getHandle(),
                ((CraftPlayer) player).getHandle(), rockSlideBlock.getState(),
                rockslideLoc.getBlockX(), rockslideLoc.getBlockY(), rockslideLoc.getBlockZ());
        return !bpe.isCancelled() && bpe.canBuild();
    }

    private boolean isRockslideCrossingOther(Location rockslideLoc, int radius) {
        for (Map.Entry<Location, Integer> otherRockslide : currentRockslides.entrySet()) {
            if (rockslideLoc.getWorld().equals(otherRockslide.getKey().getWorld())) {
                double radiusSum = radius + otherRockslide.getValue();
                if (xzDistanceSquared(rockslideLoc, otherRockslide.getKey()) <= 2 * radiusSum * radiusSum) {
                    return true;
                }
            }
        }
        return false;
    }

    private void breakTorchesBelow(Block curBlock) {
        do {
            curBlock = curBlock.getRelative(BlockFace.DOWN);

            Material curType = curBlock.getType();
            Material drop = null;
            if (curType == Material.TORCH) {
                drop = Material.TORCH;
            } else if (curType == Material.REDSTONE_TORCH_ON || curType == Material.REDSTONE_TORCH_OFF) {
                drop = Material.REDSTONE_TORCH_ON;
            }

            if (drop != null) {
                curBlock.setType(Material.AIR);
                curBlock.getWorld().dropItem(curBlock.getLocation().add(0.5, 0.5, 0.5), new ItemStack(drop));
            }
        } while (FALL_FREE_MATS.contains(curBlock.getType()));
    }

    private static double xzDistanceSquared(Location loc1, Location loc2) {
        int dx = loc1.getBlockX() - loc2.getBlockX();
        int dz = loc1.getBlockZ() - loc2.getBlockZ();
        return (dx * dx) + (dz * dz);
    }

    private class RemoveGravelTask implements Runnable {

        private final List<Block> gravelBlocks = new ArrayList<Block>();
        private final Location rockSlideLoc;

        public RemoveGravelTask(Location rockSlideLoc) {

            this.rockSlideLoc = rockSlideLoc;
        }

        public void addGravel(Block gravel) {
            gravelBlocks.add(gravel);
        }

        public boolean notEmpty() {
            return gravelBlocks.size() > 0;
        }

        @Override
        public void run() {
            for (Block gravelInitialPos : gravelBlocks) {
                Block curBlock = gravelInitialPos;
                while (true) {
                    Material curType = curBlock.getType();
                    if (curType == Material.GRAVEL) {
                        curBlock.setType(Material.AIR);
                    } else if (FALL_FREE_MATS.contains(curType)) {
                        curBlock = curBlock.getRelative(BlockFace.DOWN);
                        continue;
                    }
                    break;
                }
            }
            currentRockslides.remove(rockSlideLoc);
        }
    }
}
