package mccity.heroes.skills.turret;

import com.herocraftonline.heroes.Heroes;
import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.effects.EffectType;
import com.herocraftonline.heroes.characters.skill.PassiveSkill;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import com.herocraftonline.heroes.characters.skill.SkillType;
import com.herocraftonline.heroes.util.Messaging;
import com.herocraftonline.heroes.util.Setting;
import com.herocraftonline.heroes.util.Util;
import com.palmergames.bukkit.towny.Towny;
import com.palmergames.bukkit.towny.object.TownyUniverse;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.plugin.Plugin;

import java.util.logging.Logger;

public class SkillTurret extends PassiveSkill {

    private static Logger log;

    private static final int AI_PERIOD_TICKS = 5;

    private TurretManager turretManager;

    public static boolean useTowny = false;
    public static TownyUniverse townyUniv;

    public static Logger getLogger() {
        return log;
    }

    public SkillTurret(Heroes plugin) {
        super(plugin, "Turret");
        setDescription("You have the ability to build turret with $1 charges, which attacks enemies. Each arrow, loaded to turret adds $2 charges.");
        setIdentifiers("skill turret");
        setEffectTypes(EffectType.BENEFICIAL);
        setTypes(SkillType.SUMMON, SkillType.KNOWLEDGE);

        Bukkit.getServer().getPluginManager().registerEvents(new SkillTurretListener(), plugin);

        log = plugin.getLogger();
    }

    @Override
    public void init() {
        super.init();

        int turretsPerPlayer = Integer.parseInt(SkillConfigManager.getRaw(this, "turrets-per-player", "3"));
        boolean replaceOld = SkillConfigManager.getRaw(this, "replace-old", false);
        turretManager = new TurretManager(turretsPerPlayer, replaceOld, this);
        Messages.load(this);

        Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, new TownyCheckTask());
        Bukkit.getScheduler().scheduleSyncRepeatingTask(plugin, turretManager, 0, AI_PERIOD_TICKS);
    }

    @Override
    protected void unapply(Hero hero) {
        super.unapply(hero);
        turretManager.removeAllByHero(hero);
    }

    public ConfigurationSection getDefaultConfig() {
        ConfigurationSection defaultConfig = super.getDefaultConfig();
        defaultConfig.set("turrets-per-player", 3);
        defaultConfig.set(Setting.COOLDOWN.node(), 45000);
        defaultConfig.set("replace-old", false);
        defaultConfig.set("attack-period-ticks", 20);
        defaultConfig.set(Setting.DAMAGE.node(), 3);
        defaultConfig.set("knockback", 0);
        defaultConfig.set("fire-arrow", false);
        defaultConfig.set("initial-ammo-base", 10);
        defaultConfig.set("initial-ammo-increase", 0.4);
        defaultConfig.set("ammo-per-arrow-base", 5);
        defaultConfig.set("ammo-per-arrow-increase", 0.2);

        defaultConfig.set("lifetime.limited-lifetime", false);
        defaultConfig.set("lifetime.max-lifetime", 300000);
        defaultConfig.set("lifetime.destroy-on-owner-logout", true);
        defaultConfig.set("lifetime.destroy-on-chunk-unload", true);

        Messages.fillDefault(defaultConfig);
        return defaultConfig;
    }

    public String getDescription(Hero hero) {
        String descr = getDescription()
                .replace("$1", String.valueOf(getInitialAmmoFor(hero)))
                .replace("$2", String.valueOf(getAmmoPerArrowFor(hero)));
        StringBuilder descrSb = new StringBuilder(descr);

        int lifeTime = getTurretLifeTimeFor(hero);
        if (lifeTime != -1) {
            descrSb.append(" Lifetime: ");
            descrSb.append(lifeTime / 1000);
            descrSb.append("s.");
        }

        double cdSec = SkillConfigManager.getUseSetting(hero, this, Setting.COOLDOWN, 45000, false) / 1000.0;
        if (cdSec > 0) {
            descrSb.append(" CD:");
            descrSb.append(Util.formatDouble(cdSec));
            descrSb.append("s");
        }

        int mana = SkillConfigManager.getUseSetting(hero, this, Setting.MANA, 30, false);
        if (mana > 0) {
            descrSb.append(" M:");
            descrSb.append(mana);
        }

        return descrSb.toString();
    }

    public int getInitialAmmoFor(Hero hero) {
        return SkillConfigManager.getUseSetting(hero, this, "initial-ammo-base", 10, false) +
                (int) (SkillConfigManager.getUseSetting(hero, this, "initial-ammo-increase", 0.4, false) * hero.getSkillLevel(this));
    }

    public int getAmmoPerArrowFor(Hero hero) {
        return SkillConfigManager.getUseSetting(hero, this, "ammo-per-arrow-base", 5, false) +
                (int) (SkillConfigManager.getUseSetting(hero, this, "ammo-per-arrow-increase", 0.2, false) * hero.getSkillLevel(this));
    }

    public int getTurretLifeTimeFor(Hero hero) {
        boolean limitedLifetime = SkillConfigManager.getUseSetting(hero, this, "lifetime.limited-lifetime", false);
        if (limitedLifetime) {
            return SkillConfigManager.getUseSetting(hero, this, "lifetime.max-lifetime", 300000, false);
        }
        return -1;
    }

    public class SkillTurretListener implements Listener {

        @EventHandler(ignoreCancelled = true)
        public void onBlockPlace(BlockPlaceEvent event) {
            Block block = event.getBlockPlaced();
            if (block.getType() != Material.DISPENSER) return;
            Hero hero = plugin.getCharacterManager().getHero(event.getPlayer());
            if (!hero.hasEffect("Turret")) return;
            if (!checkTurretPlatform(block.getLocation())) return;

            if (turretManager.canAdd(hero)) {
                Long expiry = hero.getCooldown("Turret");
                if (expiry == null || (expiry <= System.currentTimeMillis())) {
                    if (turretManager.add(hero, block)) {
                        broadcast(block.getLocation(), Messages.turretPlacedBy, event.getPlayer().getName());
                        long cooldown = SkillConfigManager.getUseSetting(hero, SkillTurret.this, Setting.COOLDOWN.node(), 30000, false);
                        hero.setCooldown("Turret", System.currentTimeMillis() + cooldown);
                    }
                } else {
                    long remaining = expiry - System.currentTimeMillis();
                    Messaging.send(hero.getPlayer(), "Sorry, $1 still has $2 seconds left on cooldown!", "Turret", remaining / 1000);
                }
            } else {
                Messaging.send(hero.getPlayer(), Messages.alreadyHaveMaxTurrets);
            }
        }

        @EventHandler(ignoreCancelled = true)
        public void onBlockBreak(BlockBreakEvent event) {
            if (event.getBlock().getType() != Material.DISPENSER) return;

            Location loc = event.getBlock().getLocation();
            turretManager.baseBlockDestroyedAt(loc, event.getPlayer());
        }
    }

    private boolean checkTurretPlatform(Location dispenserLoc) {
        World world = dispenserLoc.getWorld();
        int x = dispenserLoc.getBlockX();
        int y = dispenserLoc.getBlockY();
        int z = dispenserLoc.getBlockZ();

        if (world.getBlockAt(x, y - 1, z).getType() != Material.NETHER_FENCE) return false;

        if (world.getBlockAt(x - 1 , y - 1, z).getType() != Material.NETHER_FENCE) return false;
        if (world.getBlockAt(x + 1 , y - 1, z).getType() != Material.NETHER_FENCE) return false;
        if (world.getBlockAt(x, y - 1, z - 1).getType() != Material.NETHER_FENCE) return false;
        if (world.getBlockAt(x, y - 1, z + 1).getType() != Material.NETHER_FENCE) return false;

        if (world.getBlockAt(x - 1, y, z).getType() != Material.REDSTONE_TORCH_ON) return false;
        if (world.getBlockAt(x + 1, y, z).getType() != Material.REDSTONE_TORCH_ON) return false;
        if (world.getBlockAt(x, y, z - 1).getType() != Material.REDSTONE_TORCH_ON) return false;
        if (world.getBlockAt(x, y, z + 1).getType() != Material.REDSTONE_TORCH_ON) return false;

        return true;
    }

    private class TownyCheckTask implements Runnable {

        /* Runs after all plugins had been loaded */
        @Override
        public void run() {
            Plugin towny = Bukkit.getPluginManager().getPlugin("Towny");
            if (towny != null) {
                useTowny = true;
                townyUniv = ((Towny) towny).getTownyUniverse();
                getLogger().info("Skill Turret: Using towny");
            }
        }
    }
}
