package mccity.heroes.skills.turret;

import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.util.Messaging;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.block.Block;
import org.bukkit.entity.Player;

import java.util.*;

public class TurretManager implements Runnable {

    private final Map<Location, Turret> turretMap = new HashMap<Location, Turret>();
    private final Map<String, List<Turret>> playerTurretsMap = new HashMap<String, List<Turret>>();

    private final int turretsPerPlayer;
    private final boolean replaceOld;

    private final SkillTurret skill;

    public TurretManager(int turretsPerPlayer, boolean replaceOld, SkillTurret skill) {
        this.turretsPerPlayer = turretsPerPlayer;
        this.replaceOld = replaceOld;
        this.skill = skill;
    }

    private List<Turret> getPlayerTurretList(Player player) {
        String playerName = player.getName();
        List<Turret> playerTurrets = playerTurretsMap.get(playerName);
        if (playerTurrets == null) {
            playerTurrets = new LinkedList<Turret>();
            playerTurretsMap.put(playerName, playerTurrets);
        }

        return playerTurrets;
    }

    public boolean canAdd(Hero hero) {
        List<Turret> playerTurrets = getPlayerTurretList(hero.getPlayer());
        return (replaceOld || playerTurrets.size() < turretsPerPlayer);
    }

    public boolean add(Hero hero, Block base) {
        List<Turret> playerTurrets = getPlayerTurretList(hero.getPlayer());

        boolean replace = false;
        if (playerTurrets.size() >= turretsPerPlayer) {
            if (replaceOld) {
                // remove oldest
                removeByBaseLoc(playerTurrets.get(0).getBaseBlock().getLocation());
                replace = true;
            } else {
                return false;
            }
        }

        Turret newTurret = new Turret(hero, base);
        newTurret.configure(skill);
        turretMap.put(base.getLocation(), newTurret);
        playerTurrets.add(newTurret);
        if (replace) {
            Messaging.send(hero.getPlayer(), Messages.turretReplaced);
        }
        return true;
    }

    public Turret baseBlockDestroyedAt(Location loc, Player destroyer) {
        Turret destroyed = removeByBaseLoc(loc);
        if (destroyed != null) {
            if (destroyer != null && !destroyer.getName().equals(destroyed.getOwnerName())) {
                Messaging.send(destroyed.getOwner().getPlayer(), Messages.yourTurretDestroyedBy, destroyer.getName());
                Messaging.send(destroyer, Messages.youDestroyOtherPlayerTurret, destroyed.getOwnerName());
            } else {
                Messaging.send(destroyed.getOwner().getPlayer(), Messages.yourTurretDestroyed);
            }
        }
        return destroyed;
    }

    public void removeAllByHero(Hero hero) {
        Player player = hero.getPlayer();
        List<Turret> playerTurrets = playerTurretsMap.remove(player.getName());
        if (playerTurrets != null) {
            for (Turret playerTurret : playerTurrets) {
                turretMap.remove(playerTurret.getBaseBlock().getLocation());
            }
        }
    }

    private Turret removeByBaseLoc(Location loc) {
        Turret removed = turretMap.remove(loc);
        if (removed != null) {
            List<Turret> playerTurrets = playerTurretsMap.get(removed.getOwnerName());
            for (Iterator<Turret> itr = playerTurrets.iterator(); itr.hasNext();) {
                if (itr.next() == removed) {
                    itr.remove();
                    break;
                }
            }
        }
        return removed;
    }

    @Override
    public void run() {
        if (turretMap.isEmpty()) return;

        List<Turret> forRemove = new ArrayList<Turret>();

        for (Map.Entry<Location, Turret> entry : turretMap.entrySet()) {
            Turret curTurret = entry.getValue();

            if (curTurret.isAlive()) {
                if (curTurret.getBaseBlock().getChunk().isLoaded()) {
                    curTurret.onAiTick();
                }
            } else {
                forRemove.add(curTurret);
            }
        }

        for (Turret rTurret : forRemove) {
            removeByBaseLoc(rTurret.getBaseBlock().getLocation());

            // after relogin player dont recieve message, if use Player instance before relogin, so get newest Player instance first
            Player player = Bukkit.getPlayer(rTurret.getOwnerName());
            if (player != null && player.isOnline()) {
                Messaging.send(player, Messages.yourTurretLostMagicPower);
            }
        }
    }
}
